package org.apache.stratum.resources;

/**
 * This class represent a general text resource that
 * may have been retrieved from any number of possible
 * sources.
 *
 * @author <a href="mailto:jvanzyl@apache.org">Jason van Zyl</a>
 * @author <a href="mailto:geirm@optonline.net">Geir Magnusson Jr.</a>
 * @version $Id: AbstractResource.java,v 1.1 2002/01/15 04:59:27 jvanzyl Exp $
 */
public abstract class AbstractResource
    implements Resource
{
    protected static final long MILLIS_PER_SECOND =  1000;

    /**
     * How often the file modification time is checked (in milliseconds).
     */
    protected long checkInterval = 0;

    /**
     * The file modification time (in milliseconds) for the cached template.
     */
    protected long lastModified = 0;

    /**
     * The next time the file modification time will be checked (in 
     * milliseconds).
     */
    protected long nextCheck = 0;

    /**
     *  Name of the resource
     */
    protected String name;
    
    /**
     * Encoding of this resource.
     */
    protected String encoding;

    /** 
     *  Default constructor 
     */
    public AbstractResource()
    {
    }

    /**
     * Set the name of this resource, for example
     * test.vm.
     */
    public void setName(String name)
    {
        this.name = name;
    }        

    /**
     * Get the name of this template.
     */
    public String getName()
    {
        return name;
    }        

    /**
     *  set the encoding of this resource
     *  for example, "ISO-8859-1"
     */
    public void setEncoding( String encoding )
    {
        this.encoding = encoding;
    }

    /**
     *  get the encoding of this resource
     *  for example, "ISO-8859-1"
     */
    public String getEncoding()
    {
        return encoding;
    }

    /**
     * Set the last modified time for this
     * template.
     */
    public void setLastModified(long lastModified)
    {
        this.lastModified = lastModified;
    }        

    /**
     * Return the lastModifed time of this
     * template.
     */
    public long getLastModified()
    {
        return lastModified;
    }        
    
    /**
     * Set the modification check interval.
     * @param interval The interval (in minutes).
     */
    public void setCheckInterval(long checkInterval)
    {
        this.checkInterval = checkInterval;
    }
    
    /**
     * Is it time to check to see if the resource
     * source has been updated?
     */
    public boolean requiresChecking()
    {
         // Short circuit this if checkInterval == 0
         // as this means "don't check"
        if (checkInterval <= 0 )
        {
           return false;
        }

        // see if we need to check now
        return ( System.currentTimeMillis() >= nextCheck );
    }

    /**
     * 'Touch' this template and thereby resetting
     * the nextCheck field.
     */
    public void touch()
    {
        nextCheck = System.currentTimeMillis() + 
            ( MILLIS_PER_SECOND *  checkInterval);
    }
}
